# -*- coding: utf-8 -*-
from datetime import tzinfo, timedelta, datetime

import pytz
from khayyam.helpers import force_encoded_string_output

ZERO_DELTA = timedelta(0)
TEHRAN_OFFSET = timedelta(hours=3, minutes=30)

TEHRAN_LMT_ZONE = pytz.timezone("Asia/Tehran")
TEHRAN_ZONE = datetime(2000, 1, 1).astimezone(pytz.timezone("Asia/Tehran")).tzinfo


class Timezone(tzinfo):

    def __init__(self, offset, name=None):
        assert (isinstance(offset, timedelta))
        self._offset = offset
        self._name = name
        super(Timezone, self).__init__()

    def fromutc(self, dt):
        if dt.tzinfo is None:
            raise ValueError('The object: %s is naive.' % dt)

        if dt.tzinfo != self:
            raise ValueError('Datetime timezone mismatch: %s != %s' % (dt.tzinfo, self))

        utc_offset = dt.utcoffset()
        delta = utc_offset

        if delta:
            dt += delta  # convert to standard local time

        return dt

    def utcoffset(self, dt):
        return self._offset

    def dst(self, dt):
        return ZERO_DELTA

    def __unicode__(self):
        off = self._offset
        return '%s%.2d:%.2d' % (
            '+' if off.total_seconds() >= 0 else '-',
            int(off.total_seconds() / 3600),
            int((off.total_seconds() % 3600) / 60),
        )

    __repr__ = force_encoded_string_output(__unicode__)

    def tzname(self, dt):
        if self._name:
            return self._name
        else:
            return repr(self)

    def __hash__(self):
        return hash(self._offset)


class TehranTimezone(Timezone):
    """
    Tehran timezone with a fixed +3:30 GMT offset.
    """

    def __init__(self):
        super(TehranTimezone, self).__init__(
            TEHRAN_OFFSET,
            "Asia/Tehran"
        )



import django_filters

from django_jalalify.fields import JalaliDateRangeField, JalaliDateTimeRangeField


class JalaliDateFromToRangeFilter(django_filters.DateFromToRangeFilter):
    """
    An extension on django_filters.IsoDateTimeFromToRangeFilter which accepts Jalali date as input and then
     converts it to datetime instance.
    """
    field_class = JalaliDateRangeField


class JalaliIsoDateTimeFromToRangeFilter(django_filters.IsoDateTimeFromToRangeFilter):
    """
    An extension on django_filters.IsoDateTimeFromToRangeFilter which accepts Jalali datetime as input and then
     converts it to datetime instance.
    """
    field_class = JalaliDateTimeRangeField



from django.core.exceptions import ValidationError
from django.forms import DateTimeField
from django.forms.utils import from_current_timezone
from django.utils.translation import gettext_lazy as _
from django_filters import fields, widgets

from django_jalalify import JalaliDate, JalaliDatetime


class JalaliDateTimeField(DateTimeField):
    default_error_messages = {
        "invalid": _("Enter a valid datetime of the format YYY/mm/dd HH:MM:SS."),
    }

    def to_python(self, value):
        """
        method used to convert the input value into datetime. currently, the only supported format is YYYY/mm/DD HH:MM/SS
        :param value: str
        :return: datetime. timezone aware instance
        """
        if value in self.empty_values:
            return None
        try:
            dt = JalaliDatetime.strptime(value, "%Y/%m/%d %H:%M:%S").todatetime()
        except (ValueError, TypeError):
            raise ValidationError(message={"datetime_jalali": [self.error_messages["invalid"]]}, code="invalid")
        return from_current_timezone(dt)


class JalaliDateField(DateTimeField):
    def to_python(self, value):
        """
        method used to convert the input value into date. The only supported format is YYYY/mm/DD
        :param value: str
        :return: date.
        """
        if value in self.empty_values:
            return None
        try:
            dt = JalaliDate.strptime(value, "%Y/%m/%d").todate()
        except (ValueError, TypeError):
            raise ValidationError(self.error_messages["invalid"], code="invalid")
        return dt


class JalaliDateRangeField(fields.RangeField):
    widget = widgets.DateRangeWidget

    def __init__(self, *args, **kwargs):
        fields = (
            JalaliDateField(),
            JalaliDateField(),
        )
        super().__init__(fields, *args, **kwargs)


class JalaliDateTimeRangeField(fields.RangeField):
    widget = widgets.DateRangeWidget

    def __init__(self, *args, **kwargs):
        fields = (
            JalaliDateTimeField(),
            JalaliDateTimeField(),
        )
        super().__init__(fields, *args, **kwargs)



from datetime import datetime
from typing import Tuple

import pytz

from django_jalalify import JalaliDatetime
from django_jalalify.functions import convert_date_to_int, convert_time_to_int
from django_jalalify.timezone import TEHRAN_ZONE, TEHRAN_LMT_ZONE,  TehranTimezone


def tehran_now() -> datetime: return datetime.now(pytz.timezone('Asia/Tehran'))


def get_now_tehran_jalali_datetime() -> JalaliDatetime:
    return JalaliDatetime.now(TehranTimezone())


def get_now_tehran_jalali_date_strftime(string_format="%Y/%m/%d") -> str:
    return get_now_tehran_jalali_datetime().date().strftime(string_format)


def get_now_tehran_jalali_time_strftime(string_format="%H:%M:%S") -> str:
    return get_now_tehran_jalali_datetime().time().strftime(string_format)


def get_now_tehran_jalali_date_intftime() -> int:
    """
    convert today Jalali date as a integer date
    e.g: datetime(2023-03-29) => 1402/01/09 => 14020109
    """
    today_jalali_date_strftime = get_now_tehran_jalali_date_strftime(string_format="%Y-%m-%d")
    return convert_date_to_int(today_jalali_date_strftime)


def get_now_tehran_jalali_time_intftime() -> int:
    today_jalali_time_strftime = get_now_tehran_jalali_time_strftime(string_format="%H:%M:%S")
    return convert_time_to_int(today_jalali_time_strftime)


def get_tehran_timestamp_with_three_digits_of_microsecond_accuracy(timestamp) -> str:
    return timestamp.astimezone(TEHRAN_ZONE).strftime("%Y/%m/%d %H:%M:%S:%f")[:-3]


def get_jalali_tehran_datetime_from_date_string(time) -> datetime:
    return JalaliDatetime.strptime(time, "%Y/%m/%d %H:%M:%S").replace(tzinfo=TehranTimezone).todatetime()


def int_jalali_date_to_jalali_datetime(date) -> datetime:
    """
    convert 14010105 to datetime.datetime(2022, 3, 25, 0, 0, tzinfo=+03:30 dst:60).
    """
    day = date % 100
    date //= 100
    month = date % 100
    year = date // 100
    return JalaliDatetime(year=year, month=month, day=day, tzinfo=TehranTimezone()).todatetime()


def str_of_int_to_jalali_datetime(date, time) -> datetime:
    """
    convert "14020202", "101010" to datetime.datetime(2022, 3, 25, 0, 0, tzinfo=+03:30 dst:60).
    """
    return JalaliDatetime(year=int(date[:4]), month=int(date[4:6]), day=int(date[6:8]),
                          hour=int(time[:2]), minute=int(time[2:4]), second=int(time[4:6]),
                          tzinfo=TehranTimezone()).todatetime()


def jalali_datetime_to_int(jalali_datetime) -> Tuple:
    """
    convert a JalaliDatetime to int: khayyam.JalaliDatetime(1402, 2, 3, 22, 10, 20, 155377, Yekshanbeh) -> 123456,123456
    which is the date of datetime object and the second obj is related to time
    """
    date = f"{jalali_datetime.year:02d}{jalali_datetime.month:02d}{jalali_datetime.day:02d}"
    time = f"{jalali_datetime.hour:02d}{jalali_datetime.minute:02d}{jalali_datetime.second:02d}"
    return date, time


def convert_datetime_to_custom_jalali_date(date_time: datetime) -> int:
    """
    Convert input date to Jalali integer date
    e.g: datetime(2023-03-29) => 1402/01/09 => 14020109
    """
    date_in_jalali = JalaliDatetime(date_time.astimezone(TEHRAN_LMT_ZONE)).strftime(format_string="%Y%m%d")
    return convert_date_to_int(date_in_jalali)



def convert_date_to_int(date) -> int:
    return int(str(date).replace("-", "")[:8])


def convert_time_to_int(time_string) -> int:
    return int(str(time_string).replace(":", "")[0:6])


def int_of_time_to_str(time) -> str:
    return f"{time:06d}"



from django.test import TestCase
from django.utils import timezone
from freezegun import freeze_time

from django_jalalify import JalaliDatetime
from django_jalalify.timezone import TEHRAN_ZONE, TehranTimezone


class TehranTimezoneTestCase(TestCase):

    @freeze_time("2023-5-3 10:45:36.466402", tz_offset=0)
    def test_jalali_tehran_timezone_correctness_of_dst_with_custom_tehrantimezone(self):
        must_be_jalali_in_tehran = "1402/02/13 14:15:36.466402"
        now = JalaliDatetime.now()
        now_in_jalali_and_tehran_timezone = now.astimezone(TehranTimezone()).strftime("%Y/%m/%d %H:%M:%S.%f")
        self.assertEqual(now_in_jalali_and_tehran_timezone, must_be_jalali_in_tehran)

    @freeze_time("2023-05-03 10:45:36.466402", tz_offset=0)
    def test_tehran_timezone_correctness_of_dst_with_checkout_casting(self):
        must_be_in_tehran = "2023/05/03 14:15:36.466402"
        now = timezone.now()
        now_in_tehran_timezone = now.astimezone(TEHRAN_ZONE).strftime("%Y/%m/%d %H:%M:%S.%f")
        self.assertEqual(now_in_tehran_timezone, must_be_in_tehran)

    def test_tehran_timezone_correctness_of_dst(self):
        now = timezone.now()
        now_casting_checkout_app = now.astimezone(TEHRAN_ZONE).strftime("%Y/%m/%d %H:%M:%S.%f")
        now_casting_with_custom_tehran_timezone = now.astimezone(TehranTimezone()).strftime(
            "%Y/%m/%d %H:%M:%S.%f")
        self.assertEqual(now_casting_checkout_app, now_casting_with_custom_tehran_timezone)



from typing import Union

from django_jalalify import JalaliDatetime
from django_jalalify.timezone import TEHRAN_ZONE


class FieldDateTimeInJalaliGeneratorMixin:

    def _field_datetime_in_jalali(self, field_name) -> Union[None, str]:
        field_object = self._meta.get_field(field_name)
        field_value = field_object.value_from_object(self)
        return field_value and JalaliDatetime(field_value.astimezone(TEHRAN_ZONE)).strftime(
            "%Y/%m/%d %H:%M:%S"
        )
